// ============================================================================
// CBP + FBO = MAS22 Simulation Framework
// MAS22_CircularTrack.cpp
// 30-05-22 Code Refactored (rhs included in class)
// 10-06-22 Further refactoring
// 24-06-22 New OOP to solvers - functors!
// 29-06-22 Fixup
// ============================================================================

#include "MAS22_CircularTrack.h"
#include "MAS22_Actor.h"
#include "MAS22_Structs.h"
#include "rkf45.hpp"
#include "Components/CapsuleComponent.h"
#include "Components/StaticMeshComponent.h"
#include "MAS22_Stepper.h"
#include "MAS22_Solver_RK.h"
#include "MAS22_Solver_RKF.h"

#define M1 0
#define M2 1
#define K 2
#define N 3
#define GLIDERGLIDERINTNTHETA 4

// Global 'parameters' = constants - not in modelParameters array --------
//double _angularLen;

// Global initial conditions for this model ------------------------------
// defines index into modelInitialConditions array

#define INITTHETA1 0
#define INITOMEGA1 1
#define INITTHETA2 2
#define INITOMEGA2 3

// Number of equations in rhs -------------------
#define NREQUNS 4

AMAS22_CircularTrack::AMAS22_CircularTrack()
{
	// Set this actor to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = true;

	// Need this for all actors --------------------------------------------------
	CapsuleCollider = CreateDefaultSubobject<UCapsuleComponent>("CapsuleCollider");
	CapsuleCollider->SetupAttachment(GetRootComponent());
	CapsuleCollider->SetCapsuleHalfHeight(200);
	CapsuleCollider->SetCapsuleRadius(100);
	CapsuleCollider->SetRelativeLocation(FVector(0, 0, 0));

	// Get static mesh components - the mesh name is grabbed from the content browser
	// This is specific to the model
	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset1(TEXT("/Game/CircularAirTrack/singleTrack.singleTrack"));
	UStaticMesh* Asset1 = MeshAsset1.Object;

	// Remember to place the name (here 'Base') in the header --------------------
	SingleTrack = CreateDefaultSubobject<UStaticMeshComponent>("SingleTrack");
	SingleTrack->SetStaticMesh(Asset1);
	SingleTrack->SetRelativeLocation(FVector(0, 0, 0));
	SingleTrack->SetupAttachment(CapsuleCollider);

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset2(TEXT("/Game/CircularAirTrack/outerCart.outerCart"));
	UStaticMesh* Asset2 = MeshAsset2.Object;

	SingleTrackGlider1 = CreateDefaultSubobject<UStaticMeshComponent>("SingleTrackGlider1");
	SingleTrackGlider1->SetStaticMesh(Asset2);
	SingleTrackGlider1->SetRelativeLocation(FVector(0, 0, 0));
	SingleTrackGlider1->SetupAttachment(CapsuleCollider);

	SingleTrackGlider2 = CreateDefaultSubobject<UStaticMeshComponent>("SingleTrackGlider2");
	SingleTrackGlider2->SetStaticMesh(Asset2);
	SingleTrackGlider2->SetRelativeLocation(FVector(0, 0, 0));
	SingleTrackGlider2->SetRelativeRotation(FRotator(0, 180, 0).Quaternion());
	SingleTrackGlider2->SetupAttachment(CapsuleCollider);

	SingleTrackGlider1StartRotation = SingleTrackGlider1->GetRelativeRotation();
	SingleTrackGlider2StartRotation = SingleTrackGlider2->GetRelativeRotation();


}

void AMAS22_CircularTrack::ActorInitialize() {
	Super::ActorInitialize();

	// config file stuff --------------------------
	ApplyConfig();

	// 09-07-22 Establish global logging
	InitGlobalLog();
	flog("Hi from  AMAS22_CircularTrack\n");

	u = new double[NREQUNS];
	dudt = new double[NREQUNS];
	yout = new double[NREQUNS]; // Needed for RK

	SetModelICs();
	UpdateModelICs();

	GetDefaultParameters();
	SetModelParameters();
	UpdateModelParameters();

	InitializeSolver();

	// Below must go here. Cannot go in the constructor ------------------
	solverRK = new MAS22_Solver_RK();
	solverRK->init(NREQUNS);

	solverRKF = new MAS22_Solver_RKF();

	//printf(">>>>> m1 = %f,m2 = %f, k = %f, n = %f, ggIT = %f", _m1CT, _m2CT, _kCT, _nCT, _gliderGliderIntnTheta * PI / 180.0);

	//printf("ActorInitialize()  Before create rightHandSide = %p", rightHandSide);
	rightHandSide = new rhs_CAT2(m1P, m2P, kP, nP, gliderGliderIntnThetaP*PI/180.0);
	//printf("ActorInitialize()  After create rightHandSide = %p", rightHandSide);


}

// Deal with parameters --------------------------------------------------
 
// Order must agree with defines !!
void AMAS22_CircularTrack::SetModelParameters() {
	modelParametersStruct line;
	modelParameters.Empty();

	line.val = m1P;
	line.name = "m1";
	modelParameters.Add(line);

	line.val = m2P;
	line.name = "m2";
	modelParameters.Add(line);

	line.val = kP;
	line.name = "k";
	modelParameters.Add(line);

	//printf("SetModelParameters(): k = %e", k);

	line.val = nP;
	line.name = "n";
	modelParameters.Add(line);

	line.val = gliderGliderIntnThetaP;
	line.name = "intn angle (deg)";
	modelParameters.Add(line);
}

void AMAS22_CircularTrack::UpdateModelParameters() {

	m1P = modelParameters[M1].val;
	m2P = modelParameters[M2].val;
	kP = modelParameters[K].val;
	nP = modelParameters[N].val;
	gliderGliderIntnThetaP = modelParameters[GLIDERGLIDERINTNTHETA].val;

	//printf("UpdateModelParameters(): k = %e", k);

	// - Change the params so we need a new rhs
	delete rightHandSide;
	//printf("UpdateModelParameters() Before create rightHandSide = %p", rightHandSide);
	rightHandSide = new rhs_CAT2(m1P, m2P, kP, nP, gliderGliderIntnThetaP*PI/180.0);
	//printf("UpdateModelParameters() After create rightHandSide = %p", rightHandSide);
}

void AMAS22_CircularTrack::GetDefaultParameters() {
	m1P = m1;
	m2P = m2;
	kP = k;
	nP = n;
	gliderGliderIntnThetaP = gliderGliderIntnTheta;
}

// Model initial conditions ----------------------------------------------

// Order must agree with defines
void AMAS22_CircularTrack::SetModelICs() {
	Super::SetModelICs();

	modelParametersStruct icLine;
	modelInitialConditions.Empty();			// allows for hard reset

	icLine.val = initTheta1;
	icLine.name = "initial theta 1";
	modelInitialConditions.Add(icLine);

	icLine.val = initOmega1;
	icLine.name = "initial Omega1";
	modelInitialConditions.Add(icLine);

	icLine.val = initTheta2;
	icLine.name = "initial theta 2";
	modelInitialConditions.Add(icLine);

	icLine.val = initOmega2;
	icLine.name = "initial omega 2";
	modelInitialConditions.Add(icLine);

}


void AMAS22_CircularTrack::UpdateModelICs() {
	initTheta1 = modelInitialConditions[INITTHETA1].val;
	initOmega1 = modelInitialConditions[INITOMEGA1].val;
	initTheta2 = modelInitialConditions[INITTHETA2].val;
	initOmega2 = modelInitialConditions[INITOMEGA2].val;

	// last line must be
	InitializeSolver(); // to set the ICs
}

void AMAS22_CircularTrack::InitializeSolver() {

	flag = 2;
	t = 0.0;
	time = 0;
	// Call to orig rkf code <<<<<<<<<<<<<<<<<<<<<<<<<<<<
	relerr = 10 * sqrt(r8_epsilon());
	abserr = 10 * sqrt(r8_epsilon());

	printf("AMAS22_CircularTrack::InitializeSolver() relerr = %e, abserr  = %e", relerr, abserr);

	// Set the physically determined constants ------------------------------

	u[0] = initTheta1;
	u[1] = initOmega1;
	u[2] = initTheta2;
	u[3] = initOmega2;

	// Following needed for model reset
	theta1 = u[0];
	omega1 = u[1];
	theta2 = u[2];
	omega2 = u[3];

}


void AMAS22_CircularTrack::BuildHUDLines() {
	HUDLine line;

	HUDLines.Empty();
	bHUD_exists = true;
	if (bHUD_exists) {
		Super::BuildHUDLines();
	}

	line.message = FString::Printf(TEXT("omega1 = %f"), omega1);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

	line.message = FString::Printf(TEXT("omega2 = %f"), omega2);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("theta1 = %f"), theta1);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("theta2 = %f"), theta2);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);
	line.message = FString::Printf(TEXT("thetaDiff = %f"), (theta2 - theta1));
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

}
// Following must exist for all MAS22 Actors ---------------------------------
void AMAS22_CircularTrack::LogData() {
	int m;
	m = (int(logInterval / dT));
	if (itn % (m+1) == 0) {
		dataRecords.Add(FDataRecord_CT{ time, theta1, omega1, theta2, omega2 });
	}

}
// Following must exist -------------------------------------------------------
void AMAS22_CircularTrack::ComputeDT() {

	if (bRunning) {
		LogData();
		Computation();
		Visualization(dT);
		BuildHUDLines();
		itn++;
	}
}

void AMAS22_CircularTrack::Computation() {

	// Other stuff specific to the model may go in here -----------------------

	// The following must exist -----------------------------------------------
	switch (solver) {
	case SolverEnum::EULER:
		Euler(dT);
		break;
	case SolverEnum::RUNGE_KUTTA_FEHLBERG:
		RKF(dT);
		break;
	case SolverEnum::RUNGE_KUTTA:
		RK(dT);
		break;
	default:
		break;
	}

	theta1 = u[0];
	omega1 = u[1];
	theta2 = u[2];
	omega2 = u[3];

}

// Numerical Solver Code -------------------------------------------------

void AMAS22_CircularTrack::Euler(double deltaT) {



}

void AMAS22_CircularTrack::RK(double deltaT) {
	double h = deltaT;
	flag = solverRK->step(u, dudt, time, h, yout, *rightHandSide);
	for (int i = 0; i < NREQUNS; i++) u[1] = yout[i];
	time += deltaT;

}

// Following must exist -------------------------------------------------------

void AMAS22_CircularTrack::RKF(double deltaT) {

	tout = time + deltaT;

	flag = solverRKF->r8_rkf45(NREQUNS, u,dudt, &time, tout, &relerr, abserr, flag, *rightHandSide);

	if (flag == 3) {
		printErr("Flag = 3 reset");
		flag = 2;
	}
	if (flag == 4) {
	//	printErr("Flag = 4 reset");
		flag = 2;
	}
	if (flag == 5) {
		printErr("Flag = 5 exiting");
		exit(0);
	}
	if (flag == 6) {
		printErr("Flag = 6 reset");
		flag = 2;
	}
	if (flag == 7) {
	//	printErr("Flag = 7 reset");
		flag = 2;
	}
	
	while (u[0] > 2 * M_PI) {
		u[0] -= 2 * M_PI;
	}
	while (u[0] < 0) {
		u[0] += 2 * M_PI;
	}
	while (u[2] > 2 * M_PI) {
		u[2] -= 2 * M_PI;
	}
	while (u[2] < 0) {
		u[2] += 2 * M_PI;
	}
	//printf("u[0] = %f, u[1] = %f, u[2] = %f, u[3] = %f", u[0],u[1], u[2], u[3]);
}

// Must exist - contents will depend on the model -----------------------------
void AMAS22_CircularTrack::Visualization(float deltaTime) {

	SingleTrackGlider1->SetRelativeRotation(FRotator(0, theta1 * 180 / PI, 0).Quaternion());
	SingleTrackGlider2->SetRelativeRotation(FRotator(0, theta2 * 180 / PI, 0).Quaternion());

}

void AMAS22_CircularTrack::UserInput(int key) {

}



// Console commands ------------------------------------------------------

void AMAS22_CircularTrack::SetConsoleAddit(FString name, FString strVal) {

}

void AMAS22_CircularTrack::DumpConsoleAll() {
	Super::DumpConsoleAll();

}

void AMAS22_CircularTrack::EndPlay(const EEndPlayReason::Type EndPlayReason) {
	Super::EndPlay(EndPlayReason);

	// Free stepper reserved memory
	if(solver == SolverEnum::RUNGE_KUTTA) solverRK->cleanup();

	FString logFile = FPaths::ProjectDir();
	FString fileName = "OctaveFiles/";
	fileName.Append(actorName);
	fileName.Append(".m");
	logFile.Append(fileName);

	int nrRecords = dataRecords.Num();
	FString stringToSave = "A=[\n";

	for (int i = 0; i < nrRecords; i++) {
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].time));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].theta1));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].omega1));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].theta2));
		stringToSave.Append(FString::Printf(TEXT("%e\n"), dataRecords[i].omega2));
	}

	stringToSave.Append("];\n\n");

	stringToSave.Append("time   = A(:,1);\n");
	stringToSave.Append("theta1  = A(:,2);\n");
	stringToSave.Append("omega1  = A(:,3);\n");
	stringToSave.Append("theta2  = A(:,4);\n");
	stringToSave.Append("omega2  = A(:,5);\n");

	// Use macro define in MAS22_Structs to get data into Octave file
	stringToSave.Append(logf("m1 = %f, m2 = %f\n", m1, m2));
	stringToSave.Append(logf("initTheta1 = %f, initTheta2 = %f\n", initTheta1, initTheta2));
	stringToSave.Append(logf("initOmega1 = %f, initOmega2 = %f\n", initOmega1, initOmega2));

	stringToSave.Append("figure(1)\n");
	stringToSave.Append("plot(time,(180.0/pi)*theta1,time,(180.0/pi)*theta2);\n");
	stringToSave.Append(logf("title('m1 = %f, m2 = %f');\n", m1, m2));

	stringToSave.Append("figure(2)\n");
	stringToSave.Append("plot(time,omega1,'r',time,omega2,'b');\n");
	stringToSave.Append("title('omega1 (red) and omega2 (blue)');\n");

	stringToSave.Append("figure(3)\n");
	stringToSave.Append(logf("title('m1 = %f, m2 = %f');\n", m1, m2));
	stringToSave.Append("subplot(1,2,1);\n");
	stringToSave.Append("h1 = polar(theta1,time);\n");
	stringToSave.Append("set(h1,'linewidth',3);\n");
	stringToSave.Append("subplot(1,2,2);\n");
	stringToSave.Append("h2 = polar(theta2,time);\n");
	stringToSave.Append("set(h2,'linewidth',3);\n");
	stringToSave.Append(logf("title('m1 = %f, m2 = %f');\n", m1, m2));

	stringToSave.Append("figure(4)\n");
	stringToSave.Append("plot(time,theta1-theta2);\n");
	stringToSave.Append(logf("title('theta1 - theta2: m1 = %f, m2 = %f');\n", m1, m2));

	FFileHelper::SaveStringToFile(stringToSave, *logFile);

}

void AMAS22_CircularTrack::ApplyConfig() {
	if (editorOverride == true) {
		FString configFName = FPaths::GeneratedConfigDir().Append(TEXT("MAS22_System.ini"));
		//printf("AMAS22_SkyMaster::ActorInitialize() path to config file = %s", *configFName);
		GConfig->LoadFile(configFName);
		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_CircularAirTrack"), TEXT("scaling"), scaling, configFName);
		printf("AMAS22_CircularAirTrack::ActorInitialize() >>>>>Config File read: scaling set to %f", scaling);

		GConfig->GetString(TEXT("/Script/MAS22_SciencePark.MAS22_CircularAirTrack"), TEXT("defaultSolver"), defaultSolver, configFName);
		printf("MAS22_CircularAirTrack::ActorInitialize() >>>>>Config File read: solver set to %s", *defaultSolver);
		if (defaultSolver == "RUNGE_KUTTA_FEHLBERG") solver = SolverEnum::RUNGE_KUTTA_FEHLBERG;

		GConfig->GetString(TEXT("/Script/MAS22_SciencePark.MAS22_CircularAirTrack"), TEXT("actorName"), actorName, configFName);
		printf("MAS22_CircularAirTrack::ActorInitialize() >>>>>Config File read: actor Name set to %s", *actorName);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_CircularAirTrack"), TEXT("dT"), dT, configFName);
		printf("MAS22_CircularAirTrack::ActorInitialize() >>>>>Config File read: dT set to %f", dT);

		// Default parameters ----------------------------------------------------------------------

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_CircularAirTrack"), TEXT("m1"), m1, configFName);
		printf("MAS22_CircularAirTrack::ActorInitialize() >>>>>Config File read: m1 set to %f", m1);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_CircularAirTrack"), TEXT("m2"), m2, configFName);
		printf("MAS22_CircularAirTrack::ActorInitialize() >>>>>Config File read: m2 set to %f", m2);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_CircularAirTrack"), TEXT("k"), k, configFName);
		printf("MAS22_CircularAirTrack::ActorInitialize() >>>>>Config File read: k set to %f", k);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_CircularAirTrack"), TEXT("n"), n, configFName);
		printf("MAS22_CircularAirTrack::ActorInitialize() >>>>>Config File read: n set to %f", n);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_CircularAirTrack"), TEXT("gliderGliderIntnTheta"), gliderGliderIntnTheta, configFName);
		printf("MAS22_CircularAirTrack::ActorInitialize() >>>>>Config File read: gliderGliderIntnTheta set to %f", gliderGliderIntnTheta);

	}
}
