// ============================================================================
// CBP + FBO = MAS22 Simulation Framework
// MAS22_CircularTrack.h
// 30-05-22 Code Refactored (rhs included in class)
// 10-06-22 Further refactoring
// 24-06-22 Use of functor for RHS
// 29-06-22 Fixup
// 08-07-22
// ============================================================================

#pragma once

#include "CoreMinimal.h"
#include "GameFramework/Actor.h"
#include <math.h>
#include "MAS22_Actor.h"
#include "MAS22_Stepper.h"
#include "MAS22_Structs.h"
#include "MAS22_Solver_RK.h"
#include "MAS22_Solver_RKF.h"
#include "MAS22_CircularTrack.generated.h"

// ----------------------------------------------------------------------------

// rhs Class for this actor - second attempt no params in constructor ----
class rhs_CAT2 {
public:
	// Here comes the parameter list
	double m1, m2, k, n, ggIT;
	// Constructor for this object with parameters
	rhs_CAT2(double _m1, double _m2, double _k, double _n, double _ggIT) :
		m1(_m1), m2(_m2), k(_k), n(_n), ggIT(_ggIT)
	{
		//printf("m1 = %f,m2 = %f, k = %e, n = %f, ggIT = %f", m1,m2,k,n,ggIT);
	}

	void operator() (double t, double u[], double dudt[]) {
		// Straight copy from orig rhs function
		theta1 = u[0];
		omega1 = u[1];
		theta2 = u[2];
		omega2 = u[3];

		alpha1 = 0;
		alpha2 = 0;

		while (theta1 > 2 * PI) {
			theta1 -= 2 * PI;
		}
		while (theta1 < 0) {
			theta1 += 2 * PI;
		}
		while (theta2 > 2 * PI) {
			theta2 -= 2 * PI;
		}
		while (theta2 < 0) {
			theta2 += 2 * PI;
		}

		sepn = theta2 - theta1;
		if (sepn > PI)  sepn = 2 * PI - sepn;
		if (sepn < -PI) sepn = 2 * PI + sepn;
		sepnGliders = abs(sepn);

		if ( (sepnGliders - ggIT) < 0) {
			alpha1 =  -(k / pow(sepn, n)) / m1;
			alpha2 =   (k / pow(sepn, n)) / m2;
			//printf("theta1 = %f, omega1 = %f, alpha1 = %e, theta2 = %f, omega2 = %f, alpha2 = %e sepn = %f",
			//	theta1, omega1, alpha1, theta2, omega2, alpha2,sepn);
		}

		//printf("theta1 = %f, omega1 = %f, alpha1 = %f, theta2 = %f, omega2 = %f, alpha2 = %f sepn = %f",
		//	theta1, omega1, alpha1, theta2, omega2, alpha2,sepn);

		dudt[0] = omega1;
		dudt[1] = alpha1;
		dudt[2] = omega2;
		dudt[3] = alpha2;

	}

private:
	double theta1;
	double omega1;
	double theta2;
	double omega2;

	double sepnGliders;
	double sepn;
	double alpha1;
	double alpha2;

};

// -----------------------------------------------------------------------

USTRUCT()
struct FDataRecord_CT {
	GENERATED_BODY()
		// Actual variables logged depend on specific actor
		double time, theta1, omega1, theta2, omega2;
};

UCLASS()
class MAS22_SCIENCEPARK_API AMAS22_CircularTrack : public AMAS22_Actor
{
	GENERATED_BODY()

	// Static Meshes specific to this model ---------------------------------

	UPROPERTY(VisibleDefaultsOnly)
		class UCapsuleComponent* CapsuleCollider;

	// Add static mesh components for the actor
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* SingleTrack;

	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* SingleTrackGlider1;

	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* SingleTrackGlider2;

public:

	AMAS22_CircularTrack();

private:

	// Parameters used in the .cpp file with P suffix -----------------------
	double m1P, m2P, kP, nP;
	double gliderGliderIntnThetaP;
	
	// Editable variables ---------------------------------------------------
	UPROPERTY(EditAnywhere)
		double CoRestit = 1.0;

	UPROPERTY(EditAnywhere)
		double initOmega1 = -0.5;

	UPROPERTY(EditAnywhere)
		double initTheta1 = 0;

	UPROPERTY(EditAnywhere)
		double initOmega2 = 0.0;

	UPROPERTY(EditAnywhere)
		double initTheta2 = 165*PI/180.0;

	UPROPERTY(EditAnywhere)
		double m1 = 1.0;

	UPROPERTY(EditAnywhere)
		double m2 = 2.0;

	UPROPERTY(EditAnywhere)
		double k = 1e-4;

	UPROPERTY(EditAnywhere)
		double n = 3.0;

	UPROPERTY(EditAnywhere)
		double gliderGliderIntnTheta = 14.32;

	//UPROPERTY(Config)
	//	double scaling = 100.0;

protected:

	virtual void Computation();
	virtual void Visualization(float deltaTime);
	virtual void BuildHUDLines();
	virtual void LogData();
	virtual void UserInput(int key);
	virtual void EndPlay(const EEndPlayReason::Type EndPlayReason) override;
	virtual void ApplyConfig() override;

	virtual void Euler(double dT);
	virtual void RKF(double dT);
	virtual void RK(double dT);

	// Overriden functions from MAS22_Actor ---------------------------------

	virtual void ActorInitialize() override;
	virtual void InitializeSolver() override;
	virtual void ComputeDT() override;

	virtual void SetModelParameters() override;
	virtual void UpdateModelParameters() override;
	virtual void GetDefaultParameters() override;

	virtual void SetModelICs() override;
	virtual void UpdateModelICs() override;

	virtual void DumpConsoleAll() override;

	// Console commands specific to this model ------------------------------

	virtual void SetConsoleAddit(FString name, FString strVal) override;

public:

	rhs_CAT2* rightHandSide;
	MAS22_Solver_RK* solverRK;
	MAS22_Solver_RKF* solverRKF;

private:
	double singleTrackRadius = 3.0;

	double gliderLengthUU = 64.0;
	double gliderLength;

	double theta1;
	double omega1;
	double theta2;
	double omega2;

	TArray<FDataRecord_CT> dataRecords;

	FRotator SingleTrackGlider1StartRotation;
	FRotator SingleTrackGlider2StartRotation;

};

