// =======================================================================
//
// MAS22_Stepper.h
// CBP 20-06-22
// Original stepper coding using functors
// Fourth-order Runge-Kutta
// 
// =======================================================================

#pragma once

#include "CoreMinimal.h"
#include "rkf45.hpp"

class MAS22_SCIENCEPARK_API MAS22_Stepper
{
public:
	MAS22_Stepper() {};
	MAS22_Stepper(int _nVar) {
		// Allocate temporary variables ----------------------------------
		nVar = _nVar;
		yt  = (double*)malloc(nVar * sizeof(double));
		dym = (double*)malloc(nVar * sizeof(double));
		dyt = (double*)malloc(nVar * sizeof(double));
	};
	~MAS22_Stepper() {
		// Free temporary variables on destruction -----------------------
		free(yt);
		free(dym);
		free(dyt);
	}

	void init(int _nVar) {
		nVar = _nVar;
		yt  = (double*)malloc(nVar * sizeof(double));
		dym = (double*)malloc(nVar * sizeof(double));
		dyt = (double*)malloc(nVar * sizeof(double));
	}
	void cleanup() {
		free(yt);
		free(dym);
		free(dyt);
	}

	// Classic Runge-Kutta O(4) solver --------------------------------------
	template<class D>
	int step(double y[], double* dydx, double t, double h, double* yout, D derivs) {
		double hh = h * 0.5;
		double h6 = h/6.0;
		double xh = t + hh;

		for (int i = 0; i < nVar; i++) {
			yt[i] = y[i] + hh * dydx[i];
		}
		derivs(xh, yt, dyt);

		for (int i = 0; i < nVar; i++) {
			yt[i] = y[i] + hh * dyt[i];
		}
		derivs(xh, yt, dym);

		for (int i = 0; i < nVar; i++) {
			yt[i] = y[i] + h * dym[i];
			dym[i] += dyt[i];
		}
		derivs(t + h, yt, dyt);
		for (int i = 0; i < nVar; i++) {
			yout[i] = y[i] + h6 * (dydx[i] + dyt[i] + 2.0 * dym[i]);
		}
		return 0;
	}

private:
	int nVar;
	double* yt;
	double* dym;
	double* dyt;

};


