// Fill out your copyright notice in the Description page of Project Settings.


#include "MAS22_VEH_MAG.h"
#include "MAS22_Actor.h"
#include "MAS22_Structs.h"
#include "MAS22_Solver_RK.h"
#include "MAS22_Solver_RKF.h"
#include "Components/CapsuleComponent.h"
#include "Components/StaticMeshComponent.h"


// defines for indexing into modelParameters array -----------------------

#define I_MASS 0
#define I_GRAVITY 1
#define I_C 2
#define I_K1 3
#define I_K3 4
#define I_ALPHA 5
#define I_BETA 6
#define I_RC 7
#define I_RL 8
#define I_L 9
#define I_FORCINGAMP 10
#define I_FORCINGFREQU 11

// defines index into modelInitialConditions array -------------------------

#define I_INITDISPZ 0
#define I_INITVELYZ 1

// Number of equations in rhs ----------------------------------------------
#define NREQUNS 3

// -------------------------------------------------------------------------

AMAS22_VEH_MAG::AMAS22_VEH_MAG() {

	PrimaryActorTick.bCanEverTick = false;

	// Materials ---------------------------------------------------------

	static ConstructorHelpers::FObjectFinder<UMaterial>MatAsset1(TEXT("Material'/Game/StarterContent/Materials/M_Metal_Copper.M_Metal_Copper'"));
	coilMaterial   = MatAsset1.Object;
	static ConstructorHelpers::FObjectFinder<UMaterial>MatAsset2(TEXT("Material'/Game/StarterContent/Materials/M_Metal_Chrome.M_Metal_Chrome'"));
	casingMaterial = MatAsset2.Object;
	static ConstructorHelpers::FObjectFinder<UMaterial>MatAsset3(TEXT("Material'/Game/StarterContent/Materials/M_Metal_Gold.M_Metal_Gold'"));
	magnetMaterial = MatAsset3.Object;
	static ConstructorHelpers::FObjectFinder<UMaterial>MatAsset4(TEXT("Material'/Game/StarterContent/Materials/M_Metal_Brushed_Nickel.M_Metal_Brushed_Nickel'"));
	baseMaterial = MatAsset4.Object;

	CapsuleCollider = CreateDefaultSubobject<UCapsuleComponent>("CapsuleCollider");
	CapsuleCollider->SetupAttachment(GetRootComponent());
	CapsuleCollider->SetCapsuleHalfHeight(200);
	CapsuleCollider->SetCapsuleRadius(100);
	CapsuleCollider->SetRelativeLocation(FVector(0, 0, 0));

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset1(TEXT("/Game/Monster_Truck/monsterTruckPlate.monsterTruckPlate"));
	UStaticMesh* Asset1 = MeshAsset1.Object;

	Base = CreateDefaultSubobject<UStaticMeshComponent>("Base");
	Base->SetStaticMesh(Asset1);
	Base->SetRelativeLocation(FVector(0, 0, 0));
	Base->SetRelativeScale3D(FVector(0.5, 0.5, 2.0));
	Base->SetupAttachment(CapsuleCollider);

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset2(TEXT("/Game/VEH_MAG/teleporterBase_sm.teleporterBase_sm"));
	UStaticMesh* Asset2 = MeshAsset2.Object;

	Casing = CreateDefaultSubobject<UStaticMeshComponent>("Casing");
	Casing->SetStaticMesh(Asset2);
	//Casing->SetMaterial(0, casingMaterial);
	Casing->SetRelativeLocation(FVector(0, 0, 0));
	Casing->SetRelativeScale3D(FVector(1.0, 1.0, 2.0));
	Casing->SetupAttachment(CapsuleCollider);

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset3(TEXT("/Game/VEH_MAG/Cylinder.Cylinder"));
	UStaticMesh* Asset3 = MeshAsset3.Object;

	MagTop = CreateDefaultSubobject<UStaticMeshComponent>("MagTop");
	MagTop->SetStaticMesh(Asset3);
	//MagTop->SetMaterial(0, magnetMaterial);
	MagTop->SetRelativeLocation(FVector(0, 0, 230));
	MagTop->SetRelativeScale3D(FVector(2.0, 2.0, 0.1));
	MagTop->SetupAttachment(CapsuleCollider);

	MagCent = CreateDefaultSubobject<UStaticMeshComponent>("MagCent");
	MagCent->SetStaticMesh(Asset3);
	MagCent->SetRelativeLocation(FVector(0, 0, 125));
	MagCent->SetRelativeScale3D(FVector(2.0, 2.0, 0.1));
	MagCent->SetupAttachment(CapsuleCollider);

	MagBot = CreateDefaultSubobject<UStaticMeshComponent>("MagBot");
	MagBot->SetStaticMesh(Asset3);
	MagBot->SetRelativeLocation(FVector(0, 0, 20));
	MagBot->SetRelativeScale3D(FVector(2.0, 2.0, 0.1));
	MagBot->SetupAttachment(CapsuleCollider);

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset4(TEXT("/Game/VEH_MAG/teleporterRing_sm.teleporterRing_sm"));
	UStaticMesh* Asset4 = MeshAsset4.Object;

	Coil = CreateDefaultSubobject<UStaticMeshComponent>("Coil");
	Coil->SetStaticMesh(Asset4);
	//Coil->SetMaterial(0, coilMaterial);
	Coil->SetRelativeLocation(FVector(0, 0, 125));
	Coil->SetRelativeScale3D(FVector(1.0, 1.0, 1.0));
	Coil->SetupAttachment(CapsuleCollider);

	BaseStartLocation = Base->GetRelativeLocation();
	CasingStartLocation = Casing->GetRelativeLocation();
	MagTopStartLocation = MagTop->GetRelativeLocation();
	MagBotStartLocation = MagBot->GetRelativeLocation();
	MagCentStartLocation = MagCent->GetRelativeLocation();
	CoilStartLocation = Coil->GetRelativeLocation();

}

void AMAS22_VEH_MAG::ActorInitialize() { // ------------------------------
	Super::ActorInitialize();

    Coil->SetMaterial(0, coilMaterial);
	MagTop->SetMaterial(0, magnetMaterial);
	MagBot->SetMaterial(0, magnetMaterial);
	MagCent->SetMaterial(0, magnetMaterial);
	Casing->SetMaterial(0, casingMaterial);
	Base->SetMaterial(0, baseMaterial);

	ApplyConfig();

	// Apply global logging ----------------------------------------------
	InitGlobalLog();
	flog("Hi from AMAS22_VEH_MAG\n");

	u = new double[NREQUNS];
	dudt = new double[NREQUNS];
	yout = new double[NREQUNS];

	SetModelICs();
	UpdateModelICs();

	GetDefaultParameters();
	SetModelParameters();
	UpdateModelParameters();

	InitializeSolver();

	solverRK = new MAS22_Solver_RK();
	solverRKF = new MAS22_Solver_RKF();
	rightHandSide = new rhs_VEH_MAG(_mass, _gravity, _c, _k1, _k3, _alpha, _beta, _RC, _RL, _L, _forcingAmp, _forcingFrequ);

	
	flog("mass = %f, gravity = %f\n",_mass,_gravity);

}

void AMAS22_VEH_MAG::SetModelParameters() {
	modelParametersStruct line;
	modelParameters.Empty();

	line.val = _mass;
	line.name = "mass";
	modelParameters.Add(line);

	line.val = _gravity;
	line.name = "gravity";
	modelParameters.Add(line);

	line.val = _c;
	line.name = "c";
	modelParameters.Add(line);

	line.val = _k1;
	line.name = "k1";
	modelParameters.Add(line);

	line.val = _k3;
	line.name = "k3";
	modelParameters.Add(line);

	line.val = _alpha;
	line.name = "alpha";
	modelParameters.Add(line);

	line.val = _beta;
	line.name = "beta";
	modelParameters.Add(line);

	line.val = _RC;
	line.name = "RC";
	modelParameters.Add(line);

	line.val = _RL;
	line.name = "RL";
	modelParameters.Add(line);

	line.val = _L;
	line.name = "L";
	modelParameters.Add(line);

	line.val = _forcingAmp;
	line.name = "forcing ampl";
	modelParameters.Add(line);

	line.val = _forcingFrequ;
	line.name = "forcing frequ";
	modelParameters.Add(line);

}

void AMAS22_VEH_MAG::UpdateModelParameters() {
	_mass = modelParameters[I_MASS].val;
	_gravity = modelParameters[I_GRAVITY].val;
	_c = modelParameters[I_C].val;
	_k1 = modelParameters[I_K1].val;
	_k3 = modelParameters[I_K3].val;
	_alpha = modelParameters[I_ALPHA].val;
	_beta = modelParameters[I_BETA].val;
	_RC = modelParameters[I_RC].val;
	_RL = modelParameters[I_RL].val;
	_L = modelParameters[I_L].val;
	_forcingAmp = modelParameters[I_FORCINGAMP].val;
	_forcingFrequ = modelParameters[I_FORCINGFREQU].val;

	delete rightHandSide;

	rightHandSide = new rhs_VEH_MAG(_mass, _gravity, _c, _k1, _k3, _alpha, _beta, _RC, _RL, _L, _forcingAmp, _forcingFrequ);

}
void AMAS22_VEH_MAG::GetDefaultParameters() {
	_mass = mass;
	_gravity = gravity;
	_c = c;
	_k1 = k1;
	_k3 = k3;
	_alpha = alpha;
	_beta = beta;
	_RC = RC;
	_RL = RL;
	_L = L;;
	_forcingAmp = forcingAmp;
	_forcingFrequ = forcingFrequ;

}

// Model initial conditions ----------------------------------------------

void AMAS22_VEH_MAG::SetModelICs() {
	Super::SetModelICs();

	modelParametersStruct icLine;
	modelInitialConditions.Empty();			// allows for hard reset

	icLine.val = initDispZ;
	icLine.name = "dispz";
	modelInitialConditions.Add(icLine);

	icLine.val = initVelyZ;
	icLine.name = "velyZ";
	modelInitialConditions.Add(icLine);

}



void AMAS22_VEH_MAG::UpdateModelICs() {
	initDispZ = modelInitialConditions[I_INITDISPZ].val;
	initVelyZ = modelInitialConditions[I_INITVELYZ].val;

	// last line must be
	InitializeSolver(); // to set the ICs
}


void AMAS22_VEH_MAG::InitializeSolver() {

	flag = 2;
	t = 0.0;
	time = 0;
	// Call to orig rkf code <<<<<<<<<<<<<<<<<<<<<<<<<<<<
	relerr = 10 * sqrt(solverRKF->r8_epsilon());
	abserr = 10 * sqrt(solverRKF->r8_epsilon());

	printf("AMAS22_VEH_MAG::InitializeSolver() relerr = %e, abserr  = %e", relerr, abserr);

	// Set the physically determined constants ------------------------------

	u[0] = initDispZ;
	u[1] = initVelyZ;
	u[2] = 0;

	// Following needed for model reset
	dispZ = u[0];
	velyZ = u[1];
	current = u[2];


}


void AMAS22_VEH_MAG::BuildHUDLines() {

	HUDLine line;

	HUDLines.Empty();
	bHUD_exists = true;
	if (bHUD_exists) {
		Super::BuildHUDLines();
	}

	// Add more lines specific to the model -----------------------------------

	line.message = FString::Printf(TEXT("u[0] = %f"), u[0]);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);
	//line.message = FString::Printf(TEXT("max u[0] = %f"), maxU0);
	//line.colour = VARIBLE_COLOR;
	//line.value = 0;
	//HUDLines.Add(line);
	line.message = FString::Printf(TEXT("current = %f"), u[2]);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);
	//line.message = FString::Printf(TEXT("max current = %f"), maxI);
	//line.colour = VARIBLE_COLOR;
	//line.value = 0;
	//HUDLines.Add(line);

}


// Following must exist for all MAS22 Actors ---------------------------------
void AMAS22_VEH_MAG::LogData() {
	int m;
	m = (int(logInterval / dT));
	if (itn % (m + 1) == 0) {
		dataRecords.Add(FDataRecord_VH{ time, floorPlateDispZ, dispZ, current });
	}

}
// Following must exist -------------------------------------------------------
void AMAS22_VEH_MAG::ComputeDT() {

	if (bRunning) {
		LogData();
		Computation();
		Visualization(dT);
		BuildHUDLines();
		itn++;
	}
}

void AMAS22_VEH_MAG::Computation() {

	// Other stuff specific to the model may go in here -----------------------

	// The following must exist -----------------------------------------------
	switch (solver) {
	case SolverEnum::EULER:
		Euler(dT);
		break;
	case SolverEnum::RUNGE_KUTTA_FEHLBERG:
		RKF(dT);
		break;
	case SolverEnum::RUNGE_KUTTA:
		RK(dT);
		break;
	default:
		break;
	}

	dispZ = u[0];
	velyZ = u[1];
	current = u[2];

	floorPlateDispZ = rightHandSide->getFloorPlateDispZ(time);


}

// Numerical Solver Code -------------------------------------------------

void AMAS22_VEH_MAG::Euler(double deltaT) {



}

void AMAS22_VEH_MAG::RK(double deltaT) {
	double h = deltaT;
	flag = solverRK->step(u, dudt, time, h, yout, *rightHandSide);
	for (int i = 0; i < NREQUNS; i++) u[1] = yout[i];
	time += deltaT;

}

// Following must exist -------------------------------------------------------

void AMAS22_VEH_MAG::RKF(double deltaT) {

	tout = time + deltaT;

	flag = solverRKF->r8_rkf45(NREQUNS, u, dudt, &time, tout, &relerr, abserr, flag, *rightHandSide);

	if (flag == 3) {
		printErr("Flag = 3 reset");
		flag = 2;
	}
	if (flag == 4) {
		//	printErr("Flag = 4 reset");
		flag = 2;
	}
	if (flag == 5) {
		printErr("Flag = 5 exiting");
		exit(0);
	}
	if (flag == 6) {
		printErr("Flag = 6 reset");
		flag = 2;
	}
	if (flag == 7) {
		//	printErr("Flag = 7 reset");
		flag = 2;
	}

}

void AMAS22_VEH_MAG::Visualization(float deltaTime) {

	// Move the FloorPlate
	FVector newLocation = BaseStartLocation;
	newLocation.Z += scaling * floorPlateDispZ;
	Base->SetRelativeLocation(newLocation);

	// Move the Body
	newLocation = CasingStartLocation;
	newLocation.Z += scaling * floorPlateDispZ;
	Casing->SetRelativeLocation(newLocation);

	// Move the MagTop
	newLocation = MagTopStartLocation;
	newLocation.Z += scaling * floorPlateDispZ;
	MagTop->SetRelativeLocation(newLocation);

	// Move the MagBot
	newLocation = MagBotStartLocation;
	newLocation.Z += scaling * floorPlateDispZ;
	MagBot->SetRelativeLocation(newLocation);

	// Move the Coil
	newLocation = CoilStartLocation;
	newLocation.Z += scaling * floorPlateDispZ;
	Coil->SetRelativeLocation(newLocation);

	// Move the Centre Magnet
	newLocation = MagCentStartLocation;
	newLocation.Z += scaling * dispZ;
	MagCent->SetRelativeLocation(newLocation);

}

void AMAS22_VEH_MAG::UserInput(int key) {

}

// Console commands ------------------------------------------------------

void AMAS22_VEH_MAG::SetConsoleAddit(FString name, FString strVal) {

}

void AMAS22_VEH_MAG::DumpConsoleAll() {
	Super::DumpConsoleAll();
}

// Wrap up - dump Octave files -------------------------------------------

void AMAS22_VEH_MAG::EndPlay(const EEndPlayReason::Type EndPlayReason) {
	Super::EndPlay(EndPlayReason);

	// Free stepper reserved memory
	if (solver == SolverEnum::RUNGE_KUTTA) solverRK->cleanup();

	FString logFile = FPaths::ProjectDir();
	FString fileName = "OctaveFiles/";
	fileName.Append(actorName);
	fileName.Append(".m");
	logFile.Append(fileName);

	int nrRecords = dataRecords.Num();
	FString stringToSave = "A=[\n";

	for (int i = 0; i < nrRecords; i++) {
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].time));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].floorPlateDispZ));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].dispZ));
		stringToSave.Append(FString::Printf(TEXT("%e\n"),  dataRecords[i].current));
	}

	stringToSave.Append("];\n\n");

	// Create names for use in Octave from data matrix A
	stringToSave.Append("time  = A(:,1);\n");
	stringToSave.Append("dispZ  = A(:,3);\n");
	stringToSave.Append("floorPlateDispZ  = A(:,2);\n");
	stringToSave.Append("current  = A(:,4);\n");

	// Create Octave lines for plotting
	stringToSave.Append("subplot(3,1,1)\n");
	stringToSave.Append("plot(time,floorPlateDispZ);\n");
	stringToSave.Append("title('floorPlateDispZ vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('floorPlateDispZ (m)')\n");
	stringToSave.Append("grid\n");

	stringToSave.Append("subplot(3,1,2)\n");
	stringToSave.Append("plot(time,dispZ);\n");
	stringToSave.Append("title('dispZ vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('dispZ (m)')\n");
	stringToSave.Append("grid\n");

	stringToSave.Append("subplot(3,1,3)\n");
	stringToSave.Append("plot(time,current);\n");
	stringToSave.Append("title('current vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('current (Amps)')\n");
	stringToSave.Append("grid\n");

	// Write and close file
	FFileHelper::SaveStringToFile(stringToSave, *logFile);

}

void AMAS22_VEH_MAG::ApplyConfig() {
	if (editorOverride == true) {
		FString configFName = FPaths::GeneratedConfigDir().Append(TEXT("MAS22_System.ini"));
		//printf("AMAS22_SkyMaster::ActorInitialize() path to config file = %s", *configFName);
		GConfig->LoadFile(configFName);
		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_VEH_MAG"), TEXT("scaling"), scaling, configFName);
		printf("AMAS22_VEH_MAG::ActorInitialize() >>>>>Config File read: scaling set to %f", scaling);

		GConfig->GetString(TEXT("/Script/MAS22_SciencePark.MAS22_VEH_MAG"), TEXT("defaultSolver"), defaultSolver, configFName);
		printf("MAS22_VEH_MAG::ActorInitialize() >>>>>Config File read: solver set to %s", *defaultSolver);
		if (defaultSolver == "RUNGE_KUTTA_FEHLBERG") solver = SolverEnum::RUNGE_KUTTA_FEHLBERG;

		GConfig->GetString(TEXT("/Script/MAS22_SciencePark.MAS22_VEH_MAG"), TEXT("actorName"), actorName, configFName);
		printf("MAS22_VEH_MAG::ActorInitialize() >>>>>Config File read: actor Name set to %s", *actorName);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_VEH_MAG"), TEXT("dT"), dT, configFName);
		printf("MAS22_VEH_MAG::ActorInitialize() >>>>>Config File read: dT set to %f", dT);

		// Default parameters ----------------------------------------------------------------------

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_VEH_MAG"), TEXT("mass"), mass, configFName);
		printf("MAS22_VEH_MAG::ActorInitialize() >>>>>Config File read: mass set to %f", mass);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_VEH_MAG"), TEXT("c"), c, configFName);
		printf("MAS22_VEH_MAG::ActorInitialize() >>>>>Config File read: c set to %f", c);
		
		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_VEH_MAG"), TEXT("k1"), k1, configFName);
		printf("MAS22_VEH_MAG::ActorInitialize() >>>>>Config File read: k1 set to %f", k1);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_VEH_MAG"), TEXT("k3"), k3, configFName);
		printf("MAS22_VEH_MAG::ActorInitialize() >>>>>Config File read: k3 set to %f", k3);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_VEH_MAG"), TEXT("RC"), RC, configFName);
		printf("MAS22_VEH_MAG::ActorInitialize() >>>>>Config File read: RC set to %f", RC);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_VEH_MAG"), TEXT("RL"), RL, configFName);
		printf("MAS22_VEH_MAG::ActorInitialize() >>>>>Config File read: RL set to %f", RL);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_VEH_MAG"), TEXT("L"), L, configFName);
		printf("MAS22_VEH_MAG::ActorInitialize() >>>>>Config File read: L set to %f", L);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_VEH_MAG"), TEXT("alpha"), alpha, configFName);
		printf("MAS22_VEH_MAG::ActorInitialize() >>>>>Config File read: alpha set to %f", alpha);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_VEH_MAG"), TEXT("beta"), beta, configFName);
		printf("MAS22_VEH_MAG::ActorInitialize() >>>>>Config File read: beta set to %f", beta);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_VEH_MAG"), TEXT("gravity"), gravity, configFName);
		printf("MAS22_VEH_MAG::ActorInitialize() >>>>>Config File read: gravity set to %f", gravity);

	}
}