// Fill out your copyright notice in the Description page of Project Settings.

#pragma once

#include "CoreMinimal.h"
#include "MAS22_Actor.h"
#include "MAS22_Structs.h"
#include "MAS22_Solver_RK.h"
#include "MAS22_Solver_RKF.h"
#include "MAS22_VEH_MAG.generated.h"

// rhs Class for this actor

class rhs_VEH_MAG {

public:
	// Here comes the parameter list
	double mass, gravity, c, k1, k3, alpha, beta, RC, RL, L, forcingAmp, forcingFrequ;

	// Constructor for this object with parameters
	rhs_VEH_MAG(double _mass, double _gravity, double _c, double _k1, double _k3, double _alpha,
				double _beta, double _RC, double _RL, double _L, double _forcingAmp, double _forcingFrequ) :
		mass(_mass),gravity(_gravity), c(_c), k1(_k1),k3(_k3),alpha(_alpha),beta(_beta),RC(_RC),RL(_RL),L(_L),
		forcingAmp(_forcingAmp),forcingFrequ(_forcingFrequ)
	{
		
	}

	void operator() (double t, double u[], double dudt[]) {

		floorPlateDispZ = getFloorPlateDispZ(t);
		forcing = FMath::Pow(2 * PI * forcingFrequ, 2) * floorPlateDispZ;
		
		dudt[0] = u[1] / beta;
		dudt[1] = ((-1.0 * k1 * u[0] - k3 * u[0] * u[0] * u[0] - c * u[1] - alpha * u[2]) / mass - gravity + forcing) / beta;
		dudt[2] = (alpha * u[1] - (RL + RC) * u[2]) / (L * beta);

	}

	double getFloorPlateDispZ(double t) {
		double dispZ;
		dispZ = forcingAmp * FMath::Sin(forcingFrequ * 2 * PI * t / beta);
		return dispZ;
	}


private:
	double forcing;
	double floorPlateDispZ;

};

// -----------------------------------------------------------------------

USTRUCT()
struct FDataRecord_VH {
	GENERATED_BODY()
		// Actual variables logged depend on specific actor
		double time, floorPlateDispZ, dispZ, current;
};

// -----------------------------------------------------------------------

UCLASS()
class MAS22_SCIENCEPARK_API AMAS22_VEH_MAG : public AMAS22_Actor
{
	GENERATED_BODY()
	
	UPROPERTY(VisibleDefaultsOnly)
		class UCapsuleComponent* CapsuleCollider;

	// Add static mesh components for the actor
	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Base;

	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Casing;

	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* MagTop;

	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* MagCent;

	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* MagBot;

	UPROPERTY(VisibleDefaultsOnly)
		class UStaticMeshComponent* Coil;

	// Materials ---------------------------------------------------------
	//UPROPERTY()
		class UMaterial* baseMaterial;
	//UPROPERTY()
		class UMaterial* casingMaterial;
	//UPROPERTY()
		class UMaterial* coilMaterial;
	//UPROPERTY()
		class UMaterial* magnetMaterial;

public:
	AMAS22_VEH_MAG();

private:
	// Parameters used in the cpp ----------------------------------------
	double _mass, _gravity, _c, _k1, _k3, _alpha, _beta, _RC, _RL, _L, _forcingAmp, _forcingFrequ;

public:
	// Editable variables ------------------------------------------------

	UPROPERTY(EditAnywhere, Config)
		double mass = 0.09;
	UPROPERTY(EditAnywhere, Config)
		double c = 0.054;
	UPROPERTY(EditAnywhere, Config)
		double k1 = 38.7;
	UPROPERTY(EditAnywhere, Config)
		double k3 = 160000;
	UPROPERTY(EditAnywhere, Config)
		double RC = 1200;
	UPROPERTY(EditAnywhere, Config)
		double RL = 1100;
	UPROPERTY(EditAnywhere, Config)
		double L = 1.46;
	UPROPERTY(EditAnywhere, Config)
		double alpha = 60;
	UPROPERTY(EditAnywhere, Config)
		double beta = 10;
	UPROPERTY(EditAnywhere, Config)
		double gravity = 9.8;

	UPROPERTY(EditAnywhere, Config)
		double forcingFrequ = 7.9;
	UPROPERTY(EditAnywhere, Config)
		double forcingAmp = 0.014;

	UPROPERTY(EditAnywhere, Config)
		double initDispZ = 0;
	UPROPERTY(EditAnywhere, Config)
		double initVelyZ = 0;

	// -------------------------------------------------------------------

protected:

	virtual void Computation();
	virtual void Visualization(float deltaTime);
	virtual void BuildHUDLines();
	virtual void LogData();
	virtual void UserInput(int key);
	virtual void EndPlay(const EEndPlayReason::Type EndPlayReason) override;
	virtual void ApplyConfig() override;

	virtual void Euler(double dT);
	virtual void RKF(double dT);
	virtual void RK(double dT);

	// Overriden functions from MAS22_Actor ---------------------------------

	virtual void ActorInitialize() override;
	virtual void InitializeSolver() override;
	virtual void ComputeDT() override;

	virtual void SetModelParameters() override;
	virtual void UpdateModelParameters() override;
	virtual void GetDefaultParameters() override;

	virtual void SetModelICs() override;
	virtual void UpdateModelICs() override;

	virtual void DumpConsoleAll() override;
	//virtual void InitGlobalLog() override;

	// Console commands specific to this model ------------------------------

	virtual void SetConsoleAddit(FString name, FString strVal) override;

public:

	rhs_VEH_MAG* rightHandSide;
	MAS22_Solver_RK* solverRK;
	MAS22_Solver_RKF* solverRKF;

private:
	// Model variables for cpp ---------------------------------------------
	double floorPlateDispZ, dispZ, current, velyZ;

	// Adit variables needed by the model ----------------------------------

	FVector BaseStartLocation;
	FVector CasingStartLocation;
	FVector MagTopStartLocation;
	FVector MagBotStartLocation;
	FVector MagCentStartLocation;
	FVector CoilStartLocation;

	// TArray for data logging ----------------------------------------------
	TArray<FDataRecord_VH> dataRecords;

};
