// ============================================================================
//
// ============================================================================


#include "MAS22_WindTurbine.h"
#include "MAS22_Actor.h"
#include "MAS22_Structs.h"
#include "MAS22_Solver_RK.h"
#include "MAS22_Solver_RKF.h"
#include "Components/CapsuleComponent.h"
#include "Components/StaticMeshComponent.h"
#include <chrono>
#include <thread>

//#include "DestructibleComponent.h"

#define DEBUGYELLOW false

// defines for indexing into modelParameters array -----------------------
#define I_R 0
#define I_RHO 1
#define I_J 2
#define I_TSROPT 3
#define I_CPMAX 4
#define I_REGION3KP 5
#define I_REGION3KI 6
#define I_OMEGARATED 7
#define I_REGION2BETA 8

// defines index into modelInitialConditions array -------------------------
#define I_OMEGA 0
#define I_WINDV 1

// Number of equations in rhs ----------------------------------------------
#define NREQUNS 3

// -------------------------------------------------------------------------

AMAS22_WindTurbine::AMAS22_WindTurbine() {

	PrimaryActorTick.bCanEverTick = false;

	// Main Collider
	CapsuleCollider = CreateDefaultSubobject<UCapsuleComponent>("CapsuleCollider");
	CapsuleCollider->SetupAttachment(GetRootComponent());
	CapsuleCollider->SetCapsuleHalfHeight(1000);
	CapsuleCollider->SetCapsuleRadius(100);


	// First get the mesh for the turbine tower. Dean did not do this in code.

	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset1(TEXT("StaticMesh'/Game/TurbineModels/WindTurb_Base.WindTurb_Base'"));
	UStaticMesh* Asset1 = MeshAsset1.Object;

	Tower = CreateDefaultSubobject<UStaticMeshComponent>("Tower");
	Tower->SetStaticMesh(Asset1);
	//TowerMesh->SetRelativeScale3D(FVector(.2f, .2f, .2f));
	Tower->SetupAttachment(CapsuleCollider);

	// The parent of the blades is the TowerMesh. All component rotations and translations are relative to this.

	// First get the mesh
	static ConstructorHelpers::FObjectFinder<UStaticMesh>MeshAsset2(TEXT("StaticMesh'/Game/TurbineModels/WindTurb_Blade.WindTurb_Blade'"));
	UStaticMesh* Asset2 = MeshAsset2.Object;

	Blade1 = CreateDefaultSubobject<UStaticMeshComponent>("Blade1");
	Blade1->SetStaticMesh(Asset2);
	Blade1->SetupAttachment(Tower);
	//Blade1Mesh->SetRelativeScale3D(FVector(.2f, .2f, .2f));
	Blade1->SetRelativeLocation(FVector(location.X, location.Y + 50, location.Z + 4000));

	Blade2 = CreateDefaultSubobject<UStaticMeshComponent>("Blade2");
	Blade2->SetupAttachment(Tower);
	Blade2->SetStaticMesh(Asset2);
	//Blade2Mesh->SetRelativeScale3D(FVector(.2f, .2f, .2f));
	Blade2->SetRelativeLocation(FVector(location.X, location.Y + 50, location.Z + 4000));
	FRotator bladeRot = FRotator((120), 0, 0);
	Blade2->SetRelativeRotation(bladeRot);

	Blade3 = CreateDefaultSubobject<UStaticMeshComponent>("Blade3");
	Blade3->SetupAttachment(Tower);
	Blade3->SetStaticMesh(Asset2);
	//Blade3Mesh->SetRelativeScale3D(FVector(.2f, .2f, .2f));
	Blade3->SetRelativeLocation(FVector(location.X, location.Y + 50, location.Z + 4000));
	bladeRot = FRotator((240), 0, 0);
	Blade3->SetRelativeRotation(bladeRot);

}

void AMAS22_WindTurbine::ActorInitialize() { // ---------------------------
	Super::ActorInitialize();

	if (DEBUGYELLOW) printYell("AMAS22_WindTurbine::ActorInitialize()");

	// Call to function to apply the config parameters -------------------
	ApplyConfig();

	// Apply global logging ----------------------------------------------
	InitGlobalLog();
	flog("Hi from AMAS22_WindTurbine\n");

	u = new double[NREQUNS];
	dudt = new double[NREQUNS];
	yout = new double[NREQUNS]; // Required for Runge-Kutta


	SetModelICs();
	UpdateModelICs();

	GetDefaultParameters();
	SetModelParameters();
	UpdateModelParameters();

	InitializeSolver();

	// Create the solver object and the right-hand-side Functor ----------
	solverRK = new MAS22_Solver_RK();
	//solverRK->init(3);
	solverRKF = new MAS22_Solver_RKF();
	// -------------------------------------------------------------------
	
	// Fill out param list. Precede each by underscore
	rightHandSide = new rhs_WindTurbine(_R, _rho, _J, _tsrOpt, _cpMax, _region3Kp, _region3Ki, _omegaRated, _region2Beta);

	// load CplArray
	rightHandSide->ReadCpLambdaFile();

	// Dump out CplArray
	dumpCplArray(FString("In ActorInitialize()"));

	stepTime = 0;

	double v = rightHandSide->getWindV();
	//printf("windV set = %f", v); rightHandSide->setWindSpeed(windVInit);

	//Testing
	tsr = 3.0;
	beta = 3.3;
	double cp = rightHandSide->LookUpInCpLambdaArray(tsr, beta);
	if (DEBUGYELLOW) printYell("AMAS22_WindTurbine::ActorInitialize() Test tsr = %f, beta = %f, Cp = %f", tsr, beta, cp);

	tsr = 4.0;
	beta = 3.3;
	cp = rightHandSide->LookUpInCpLambdaArray(tsr, beta);
	if (DEBUGYELLOW) printYell("AMAS22_WindTurbine::ActorInitialize() Test tsr = %f, beta = %f, Cp = %f", tsr, beta, cp);

	// Testing ground taking set number of steps ---------------------------------
	//printYell("rhsPointer  = %p", rightHandSide);
	//printYell("dT = %f", dT);
	//for (int i = 0; i < NREQUNS; i++) printYell("u[%d] = %f", i, u[i]);
	time = 0;
	int nrSteps = 1;

	double localpower;
	localpower = rightHandSide->getPower(666);
	//printErr(">>>>>>>>>>>>>  actor init 1 power = %f", localpower);

	for (int step = 0; step < nrSteps; step++) {
		tout = time + dT;
		flag = 1;
		flag = solverRKF->r8_rkf45(NREQUNS, u, dudt, &time, tout, &relerr, abserr, flag, *rightHandSide);

		localpower = rightHandSide->getPower(666);
		printErr(">>>>>>>>>>>>>  actor init 1 power = %f", localpower);

	}

	// Testing how to calculate power in cpp -----------------------------
	
	double localtsr = windVInit * u[1] / _R;
	double localCp;
	double localPower;
	localCp = rightHandSide->LookUpInCpLambdaArray(localtsr, u[2]);
	localPower = 0.5 * _rho * PI * pow(_R, 5) * localCp * pow(u[1], 3);
	//printYell("localtsr = %f, localCp  = %f, localPower = %f", localtsr, localCp, localPower);

	// Must set windV -------------------------------------------------------
	windV = windVInit;

}

void AMAS22_WindTurbine::dumpCplArray(FString message) {
	TArray<FCplArray> array;
	rightHandSide->getCplArray(&array);
	flog("%s\n", *message);
	//printYell("%s\n", *message);
	//printYell("Array size %d", array.Num());
	flog("Array size %d\n", array.Num());
	for (int row = 0; row < 91; row++) {
		for (int col = 0; col < 13; col++) {
			flog("[%d][%d] %f", row, col, array[row][col]);
		}
		flog("\n");
	}
}

void AMAS22_WindTurbine::SetModelParameters() { //-------------------------
	modelParametersStruct line;
	if (DEBUGYELLOW) printYell("AMAS22_WindTurbine::SetModelParameters()");
	modelParameters.Empty();

	line.val = _R;
	line.name = "R";
	modelParameters.Add(line);

	line.val = _rho;
	line.name = "rho";
	modelParameters.Add(line);

	line.val = _J;
	line.name = "J";
	modelParameters.Add(line);

	line.val = _tsrOpt;
	line.name = "tsrOpt";
	modelParameters.Add(line);

	line.val = _cpMax;
	line.name = "cpMax";
	modelParameters.Add(line);

	line.val = _region3Kp;
	line.name = "region3Kp";
	modelParameters.Add(line);

	line.val = _region3Ki;
	line.name = "region3Ki";
	modelParameters.Add(line);

	line.val = _omegaRated;
	line.name = "omegaRated";
	modelParameters.Add(line);

	line.val = _region2Beta;
	line.name = "region2Beta";
	modelParameters.Add(line);

}

void AMAS22_WindTurbine::UpdateModelParameters() { // ---------------------
	if (DEBUGYELLOW) printYell("AMAS22_WindTurbine::UpdateModelParameters()");
	// Add lines for all parameters
	_R = modelParameters[I_R].val;
	_rho = modelParameters[I_RHO].val;
	_J = modelParameters[I_J].val;
	_tsrOpt = modelParameters[I_TSROPT].val;
	_cpMax = modelParameters[I_CPMAX].val;
	_region3Kp = modelParameters[I_REGION3KP].val;
	_region3Ki = modelParameters[I_REGION3KI].val;
	_omegaRated = modelParameters[I_OMEGARATED].val;
	_region2Beta = modelParameters[I_REGION2BETA].val;

	// Fill out param list. Precede each by underscore
	delete rightHandSide;
	rightHandSide = new rhs_WindTurbine(_R, _rho, _J, _tsrOpt, _cpMax, _region3Kp, _region3Ki, _omegaRated, _region2Beta);

	//printYell("rhsPointer  = %p", rightHandSide);

	// load CplArray
	rightHandSide->ReadCpLambdaFile();

	// Dump out CplArray
	dumpCplArray(FString("In UpdateModelParameters()"));
	// Copy and paste from line in ActorInitialize();

}
void AMAS22_WindTurbine::GetDefaultParameters() { // ----------------------
	if (DEBUGYELLOW) printYell("AMAS22_WindTurbine::GetDefaultParameters()");
	// Add lines for all parameters
	_R = R;
	_rho = rho;
	_J = J;
	_tsrOpt = tsrOpt;
	_cpMax = cpMax;
	_region3Kp = region3Kp;
	_region3Ki = region3Ki;
	_omegaRated = omegaRated;
	_region2Beta = region2Beta;

}

// Model initial conditions ----------------------------------------------

void AMAS22_WindTurbine::SetModelICs() {
	if (DEBUGYELLOW) printYell("AMAS22_WindTurbine::SetModelICs()");
	Super::SetModelICs();

	modelParametersStruct icLine;
	modelInitialConditions.Empty();			// allows for hard reset

	icLine.val = omegaInit;
	icLine.name = "omega";
	modelInitialConditions.Add(icLine);

	icLine.val = windVInit;
	icLine.name = "windV";
	modelInitialConditions.Add(icLine);

}


void AMAS22_WindTurbine::UpdateModelICs() { // ----------------------------
	if (DEBUGYELLOW) printYell("AMAS22_WindTurbine::UpdateModelICs()");
	// Add lines for all initial conditions
	omegaInit = modelInitialConditions[I_OMEGA].val;
	windVInit = modelInitialConditions[I_WINDV].val;

	// last line must be
	InitializeSolver(); // to set the ICs
}


void AMAS22_WindTurbine::InitializeSolver() { // --------------------------
	if (DEBUGYELLOW) printYell("AMAS22_WindTurbine::InitializeSolver()");

	flag = 2;
	t = 0.0;
	time = 0;
	itn = 0;
	// Call to orig rkf code <<<<<<<<<<<<<<<<<<<<<<<<<<<<

	// Increased by a factor of 10 to avoid flag = 4 issues. Needs thought
	relerr = 1000 * sqrt(solverRKF->r8_epsilon());
	abserr = 1000 * sqrt(solverRKF->r8_epsilon());

	printf("AMAS22_WindTurbine::InitializeSolver() relerr = %e, abserr  = %e", relerr, abserr);

	// Set the physically determined constants ------------------------------
	// Apply your initial conditions to the state vector u[...]
	u[0] = 0;
	u[1] = omegaInit;
	u[2] = region2Beta;

	// Grab the state values and copy into your system variabkes
	// (needed by model reset)
	theta = u[0];
	omega = u[1];
	beta = u[2];
	
	//for (int i = 0; i < NREQUNS; i++) printYell("initializeSolver():  u[%d] = %f", i, u[i]);
}

void AMAS22_WindTurbine::BuildHUDLines() { // -----------------------------

	//if (DEBUGYELLOW) printYell("AMAS22_WindTurbine::BuildHUDLines");
	HUDLine line;

	HUDLines.Empty();
	bHUD_exists = true;
	if (bHUD_exists) {
		Super::BuildHUDLines();
	}

	// Add more lines specific to the model -----------------------------------

	line.message = FString::Printf(TEXT("windV = %f"), windV);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

	line.message = FString::Printf(TEXT("region %d"), region);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

	line.message = FString::Printf(TEXT("wind type %d"), windType);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

	line.message = FString::Printf(TEXT("beta = %f"), beta);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

	line.message = FString::Printf(TEXT("tsr = %f"), tsr);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

	line.message = FString::Printf(TEXT("Cp = %f"), Cp);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

	line.message = FString::Printf(TEXT("omega = %f"), omega);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

	line.message = FString::Printf(TEXT("power = %f"), power);
	line.colour = VARIBLE_COLOR;
	line.value = 0;
	HUDLines.Add(line);

}


// Following must exist for all MAS22 Actors ---------------------------------
void AMAS22_WindTurbine::LogData() {
	int m;
	m = (int(logInterval / dT));
	if (itn % (m + 1) == 0) {
		dataRecords.Add(FDataRecord_WindT{ time, theta, omega, windV, tsr, Cp, beta, power });
		// In above complete the list of your variables to log
	}

}
// Following must exist -------------------------------------------------------
void AMAS22_WindTurbine::ComputeDT() {

	if (bRunning) {
		LogData();
		Computation();
		Visualization(dT);
		BuildHUDLines();
		itn++;
	}
}

void AMAS22_WindTurbine::Computation() { // -------------------------------

	// The following must exist ------------------------------------------

	switch (solver) {
	case SolverEnum::EULER:
		Euler(dT);
		break;
	case SolverEnum::RUNGE_KUTTA_FEHLBERG:
		RKF(dT);
		break;
	case SolverEnum::RUNGE_KUTTA:
		RK(dT);
		break;
	default:
		break;
	}

	// Grab the state vector values into your model values --------
	theta = u[0];
	omega = u[1];
	beta = u[2];
	//omega = dudt[0];

	tsr = omega * R / windV;
	currentCp = rightHandSide->getCp();

	// Set value of wind speed based on wind type -----------------
	switch (windType) {
	case stepped:
		windV = windV + steppedWindSpeed();
		//printYell(">>> windV = %f", windV);
		break;

	case series:
		int m;
		m = (int(changeInterval / dT));
		if (itn % (m + 1) == 0) {
			windV = seriesWindSpeed(averageWindSpeed);
		}
		
		break;

	case steady:
		windV = windVInit;
		break;
	}

	//printf("windV = %f", windV);

	rightHandSide->setWindSpeed(windV);

	// Update the region ------------------------------------------

	if (windV < vCutIn && region != 4) {
		region = 1;
	}
	else if (windV > vCutOut) {
		region = 4;
		printErr("INTO REGIONS 4: windV = %f", windV);
	}
	else if (windV > vCutIn && omega < omegaRated && region != 4) {
		region = 2;
	}
	else if(region != 4) {
		region = 3;
	}

	rightHandSide->setRegion(region);
	

	// Try to calculate power here ---------------------------------------

	double localtsr = _R * u[1] / windV;
	double localCp;
	double localPower;

	localCp = rightHandSide->LookUpInCpLambdaArray(localtsr, u[2]);
	
	// fudge the region 4 behaviour. Need to artificially drop Cp -----------
	if (region != 4) {
		savedCp = localCp;
		savedTSR = localtsr;		
	} else {
		localCp = localtsr * savedCp / savedTSR;
	}
	// end of fudge ---------------------------------------------------------

	// Following is the power in the wind
	localPower = 0.5 * _rho * PI * pow(_R, 5) * localCp * pow(u[1], 3) / pow(localtsr, 3);

	// Let's calc the generator output power
	double torqueG = (0.5 * rho * M_PI * pow(R, 5) * (cpMax / pow(tsrOpt, 3)) * pow(omega, 2));
	localPower = torqueG * omega;

	// clean this up later --------------------------------------------------
	power = localPower;
	Cp = localCp;
	

}


double AMAS22_WindTurbine::steppedWindSpeed() { // --------------------------
	double additWindV;
	if (stepTime >= windStepInterval) {
		stepTime = 0;
		additWindV = windStepHeight;
	}

	stepTime += dT;
	return additWindV;
}

//calculate rayleigh wind speed
double AMAS22_WindTurbine::seriesWindSpeed(double averageSpeed) {
	double test;
	double calcSpeed;
	double c;

	c = 2 * averageSpeed / FMath::Sqrt(M_PI);
	test = FMath::RandRange(0.0f, 1.0f);
	calcSpeed = c * FMath::Sqrt(-1 * FMath::Loge(test));

	if (calcSpeed >= vCutOut) {
		calcSpeed = vCutOut*0.9;
	}
	//printf("calcSpeed = %f", calcSpeed);
	return calcSpeed;
}

// Numerical Solver Code -------------------------------------------------

void AMAS22_WindTurbine::Euler(double deltaT) { // ------------------------


}

void AMAS22_WindTurbine::RK(double deltaT) { // ---------------------------
	double h = deltaT;
	flag = solverRK->step(u, dudt, time, h, yout, *rightHandSide);
	for (int i = 0; i < NREQUNS; i++) u[1] = yout[i];
	time += deltaT;

}


void AMAS22_WindTurbine::RKF(double deltaT) { // -------------------------

	tout = time + dT;
	flag = solverRKF->r8_rkf45(NREQUNS, u, dudt, &time, tout, &relerr, abserr, flag, *rightHandSide);
	theta = u[0];

	if (flag == 7) {
		printYell("BLX:: Warning - RKF flag = 7 reset at time %f to flag = 2",time);
		flag = 2;
	}
	
	
}

void AMAS22_WindTurbine::Visualization(float deltaTime) { // --------------

	double totalRot;
	totalRot = 180*theta/PI;

	Blade1->SetRelativeRotation(FRotator(totalRot, 0, 0).Quaternion());
	Blade2->SetRelativeRotation(FRotator(totalRot + 120, 0, 0).Quaternion());
	Blade3->SetRelativeRotation(FRotator(totalRot + 240, 0, 0).Quaternion());



}

void AMAS22_WindTurbine::UserInput(int key) { // --------------------------

}

// Console commands ------------------------------------------------------

void AMAS22_WindTurbine::SetConsoleAddit(FString name, FString strVal) {

}

void AMAS22_WindTurbine::DumpConsoleAll() { // ----------------------------
	Super::DumpConsoleAll();
}

// Wrap up - dump Octave files -------------------------------------------

void AMAS22_WindTurbine::EndPlay(const EEndPlayReason::Type EndPlayReason) {
	Super::EndPlay(EndPlayReason);

	// Free stepper reserved memory
	if (solver == SolverEnum::RUNGE_KUTTA) solverRK->cleanup();

	FString logFile = FPaths::ProjectDir();
	FString fileName = "OctaveFiles/";
	fileName.Append(actorName);
	fileName.Append(".m");
	logFile.Append(fileName);

	int nrRecords = dataRecords.Num();
	FString stringToSave = "A=[\n";

	for (int i = 0; i < nrRecords; i++) {
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].time));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].windV));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].omega));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].beta));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].tsr));
		stringToSave.Append(FString::Printf(TEXT("%e,  "), dataRecords[i].Cp));
		stringToSave.Append(FString::Printf(TEXT("%e \n"), dataRecords[i].power));
	}

	stringToSave.Append("];\n\n");

	// Create names for use in Octave from data matrix A
	stringToSave.Append("time   = A(:,1);\n");
	stringToSave.Append("windV  = A(:,2);\n");
	stringToSave.Append("omega  = A(:,3);\n");
	stringToSave.Append("beta   = A(:,4);\n");
	stringToSave.Append("tsr    = A(:,5);\n");
	stringToSave.Append("Cp     = A(:,6);\n");
	stringToSave.Append("power  = A(:,7);\n");

	// Create Octave lines for plotting
	stringToSave.Append("subplot(6,1,1);\n");
	stringToSave.Append("plot(time,windV,'linewidth',2);\n");
	stringToSave.Append("title('windV vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('windV (m/sec)')\n");
	stringToSave.Append("line([0,max(time)],[11.7,11.7]);\n");
	stringToSave.Append("grid\n");

	stringToSave.Append("subplot(6,1,2);\n");
	stringToSave.Append("plot(time,omega,'linewidth',2);\n");
	stringToSave.Append("title('omega vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('omega (rad/sec)')\n");
	stringToSave.Append("line([0,max(time)],[3.86,3.86]);\n");
	stringToSave.Append("grid\n");

	stringToSave.Append("subplot(6,1,3);\n");
	stringToSave.Append("plot(time,beta,'linewidth',2);\n");
	stringToSave.Append("title('beta vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('beta (degs) ')\n");
	stringToSave.Append("grid\n");

	stringToSave.Append("subplot(6,1,4);\n");
	stringToSave.Append("plot(time,tsr,'linewidth',2);\n");
	stringToSave.Append("title('tsr vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('tsr ')\n");
	stringToSave.Append("line([0,max(time)],[6.6,6.6]);\n");
	stringToSave.Append("grid\n");

	stringToSave.Append("subplot(6,1,5);\n");
	stringToSave.Append("plot(time,Cp,'linewidth',2);\n");
	stringToSave.Append("title('Cp vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('Cp ')\n");
	stringToSave.Append("line([0,max(time)],[0.463,0.463]);\n");
	stringToSave.Append("grid\n");

	stringToSave.Append("subplot(6,1,6);\n");
	stringToSave.Append("plot(time,power/1000,'linewidth',2);\n");
	stringToSave.Append("title('power vs. time')\n");
	stringToSave.Append("xlabel('time (secs)')\n");
	stringToSave.Append("ylabel('power (kW) ')\n");
	stringToSave.Append("line([0,max(time)],[600,600]);\n");
	stringToSave.Append("grid\n");

	// Write and close file
	FFileHelper::SaveStringToFile(stringToSave, *logFile);

}

void AMAS22_WindTurbine::ApplyConfig() { // -------------------------------
	if (editorOverride == true) {
		FString configFName = FPaths::GeneratedConfigDir().Append(TEXT("MAS22_System.ini"));
		//printf("AMAS22_SkyMaster::ActorInitialize() path to config file = %s", *configFName);
		GConfig->LoadFile(configFName);
		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_WindTurbine"), TEXT("scaling"), scaling, configFName);
		printf("AMAS22_WindTurbine::ActorInitialize() >>>>>Config File read: scaling set to %f", scaling);

		GConfig->GetString(TEXT("/Script/MAS22_SciencePark.MAS22_WindTurbine"), TEXT("defaultSolver"), defaultSolver, configFName);
		printf("MAS22_WindTurbine::ActorInitialize() >>>>>Config File read: solver set to %s", *defaultSolver);
		if (defaultSolver == "RUNGE_KUTTA_FEHLBERG") solver = SolverEnum::RUNGE_KUTTA_FEHLBERG;

		GConfig->GetString(TEXT("/Script/MAS22_SciencePark.MAS22_WindTurbine"), TEXT("actorName"), actorName, configFName);
		printf("MAS22_WindTurbine::ActorInitialize() >>>>>Config File read: actor Name set to %s", *actorName);

		GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_WindTurbine"), TEXT("dT"), dT, configFName);
		printf("MAS22_WindTurbine::ActorInitialize() >>>>>Config File read: dT set to %f", dT);

		// Default parameters ----------------------------------------------------------------------

		//GConfig->GetDouble(TEXT("/Script/MAS22_SciencePark.MAS22_WindTurbine"), TEXT("mass"), mass, configFName);
		//printf("MAS22_WindTurbine::ActorInitialize() >>>>>Config File read: mass set to %f", mass);



	}
}