
// =======================================================================
//
// MAS22_WindTurbine.h
// 
// CBP 18-08-22
//
// =======================================================================

#pragma once

#include "CoreMinimal.h"
#include "MAS22_Actor.h"
#include "MAS22_Structs.h"
#include "MAS22_Solver_RK.h"
#include "MAS22_Solver_RKF.h"

#include "omp.h"

#include "MAS22_WindTurbine.generated.h"

#define DEBUGYELLOW true

//the struct for the 2d array needed for the cpl

USTRUCT()
struct FCplArray {
	GENERATED_BODY()
public:
	TArray<double> cplVal;
	double operator[] (int32 i) {
		return cplVal[i];
	}
	void Add(double value) {
		cplVal.Add(value);
	}
};



class rhs_WindTurbine { // ------------------------------------------------

public:
	// Here comes the parameter list
	double R, rho, J, tsrOpt, cpMax, region3Kp, region3Ki, omegaRated, region2Beta;

	// Constructor for this object with parameters
	rhs_WindTurbine(double _R, double _rho, double _J, double _tsrOpt, double _cpMax, double _region3Kp, double _region3Ki,
		double _omegaRated, double _region2Beta) :
		R(_R), rho(_rho), J(_J), tsrOpt(_tsrOpt), cpMax(_cpMax), region3Kp(_region3Kp), region3Ki(_region3Ki), 
		omegaRated(_omegaRated), region2Beta(_region2Beta)
	{
		//if (DEBUGYELLOW) printYell("rhs_WindTurbine::Constructor J = %f",J);
	}

	void operator() (double t, double u[], double dudt[]) {
		

		// Following prevents tsr from going to zero and causing overflow
		if (region == 1) {
			u[1] = 1e-5;
		}

		omega = u[1];
		beta = u[2];

		tsr = omega * R / windV;

		Cp = LookUpInCpLambdaArray(tsr, beta);

		torqueW = (0.5 * rho * M_PI * pow(R, 5) * (Cp / pow(tsr, 3)) * pow(omega, 2));
		if (tsr == 0) torqueW = 0;

		torqueG = (0.5 * rho * M_PI * pow(R, 5) * (cpMax / pow(tsrOpt, 3)) * pow(omega, 2));

		dudt[0] = omega;
		dudt[1] = (torqueW - torqueG) / J;
		dudt[2] = 0.0;

		if (region == 3) {
			dudt[2] = -region3Ki * (omegaRated - omega) + region3Kp * dudt[1];
		}
		else if(region == 2) {
			dudt[2] = 10.0 * (region2Beta - beta);
		}
		else if (region == 4) {
			dudt[1] = -0.5*u[1];
			dudt[2] = 10.0 * (region2Beta - beta);
			//printOrange("Region = 4")
		}
		
		//normalise
		if (u[0] > 2*M_PI)  u[0] -= 2 * M_PI;

	}

	// Variables for this class
private:
	volatile double tsr, Cp;

	double theta, omega, beta;
	double torqueW, torqueG;// error;
	double windV;
	TArray<FCplArray> CplArray;
	int region;
	double power1 = 0.11115;

	// Functions called from cpp

public:
	double getR() {
		return R;
	}

	void setRegion(int _region) {
		region = _region;
		//printOrange("rhs region set to %d", region);
	}

	void setWindSpeed(double _windV) {
		windV = _windV;
		//printOrange("rhs windV set to %f", windV);
	}
	double getWindV() {
		return windV;
	}
	// none of the following gets work

	double getTSR() {
		return tsr;
	}

	double getCp() {
		return Cp;
	}

	void setPower(double source, double _power) {
		power1 = _power;
		//printOrange(">>>>>>>>>>>>> rhs setPower() %f  = %f", source, power1);
	//	getPower(111);
	}
    void getPower( volatile double* _power) {
		_power = &power1;
	  	//printOrange(">>>>>>>>>>>>> rhs getPower() 1 = %f", power1);
	}
	double getPower(double source) {
		//printOrange(">>>>>>>>>>>>> rhs getPower() %f = %f", source, power1);
		//return 111.0;	// this works
		return power1;
	}

	//populate the cpl array with data from external text document

	void ReadCpLambdaFile() {
		//printYell("rhs_WindTurbine::ReadCpLambdaFile()");
		//import data from txt file
		FString importedString;
		FString file = FPaths::ProjectConfigDir();
		file.Append(TEXT("CpLambda.txt"));
		FFileHelper::LoadFileToString(importedString, *file);

		//split by lines and commas and add data to 2d array
		TArray<FString> parsedString;
		importedString.ParseIntoArray(parsedString, TEXT("\n"), true);
		for (int i = 0; i < parsedString.Num(); i++) {
			CplArray.Add(FCplArray());
			TArray<FString> parsedLine;
			FString line = parsedString[i];
			line.ParseIntoArray(parsedLine, TEXT(","), true);
			for (int j = 0; j < parsedLine.Num(); j++) {
				CplArray[i].Add(FCString::Atod(*parsedLine[j]));
				//printYell("CplArray[%d][%d] = %f",i,j, CplArray[i][j]);
			}
		}
		//printYell("ReadCpLambdaFile() CplArray size = %d", CplArray.Num());
	}
	
	void getCplArray(TArray<FCplArray>* array) {
		*array = CplArray;
	}

	double LookUpInCpLambdaArray(double tsr_, double pitch_) {
		int row, col;
		double b1, b2, Cp1, Cp2;
		double Cpp;
		double b1_curr, b2_curr;

		//printYell("rhs_WindTurbine::LookUpInCpLambdaArray\n");
		row = (int)((tsr_ - 3.0) * 10);
		if (row > 90) row = 90;
		if (row < 0) row = 0;

		// catch the limits of the pitch_
		if (pitch_ < 3.2) col = 0;
		if (pitch_ > 33.7) col = 12;

		//printYell("row = %d\n", row);

		// now do the interpolations.

		b1 = 3.2;
		b2 = 3.7;
		if ((pitch_ > b1) && (pitch_ <= b2)) {
			col = 0;
			b1_curr = b1;
			b2_curr = b2;
		}

		b1 = 3.7;
		b2 = 4.2;
		if ((pitch_ > b1) && (pitch_ <= b2)) {
			col = 1;
			b1_curr = b1;
			b2_curr = b2;
		}

		b1 = 4.2;
		b2 = 4.7;
		if ((pitch_ > b1) && (pitch_ <= b2)) {
			col = 2;
			b1_curr = b1;
			b2_curr = b2;
		}

		b1 = 4.7;
		b2 = 5.7;
		if ((pitch_ > b1) && (pitch_ <= b2)) {
			col = 3;
			b1_curr = b1;
			b2_curr = b2;
		}

		b1 = 5.7;
		b2 = 6.7;
		if ((pitch_ > b1) && (pitch_ <= b2)) {
			col = 4;
			b1_curr = b1;
			b2_curr = b2;
		}

		b1 = 6.7;
		b2 = 7.7;
		if ((pitch_ > b1) && (pitch_ <= b2)) {
			col = 5;
			b1_curr = b1;
			b2_curr = b2;
		}

		b1 = 7.7;
		b2 = 8.7;
		if ((pitch_ > b1) && (pitch_ <= b2)) {
			col = 6;
			b1_curr = b1;
			b2_curr = b2;
		}

		b1 = 8.7;
		b2 = 13.7;
		if ((pitch_ > b1) && (pitch_ <= b2)) {
			col = 7;
			b1_curr = b1;
			b2_curr = b2;
		}

		b1 = 13.7;
		b2 = 18.7;
		if ((pitch_ > b1) && (pitch_ <= b2)) {
			col = 8;
			b1_curr = b1;
			b2_curr = b2;
		}

		b1 = 18.7;
		b2 = 23.7;
		if ((pitch_ > b1) && (pitch_ <= b2)) {
			col = 9;
			b1_curr = b1;
			b2_curr = b2;
		}

		b1 = 23.7;
		b2 = 28.7;
		if ((pitch_ > b1) && (pitch_ <= b2)) {
			col = 10;
			b1_curr = b1;
			b2_curr = b2;
		}

		b1 = 28.7;
		b2 = 33.7;
		if ((pitch_ > b1) && (pitch_ <= b2)) {
			col = 11;
			b1_curr = b1;
			b2_curr = b2;
		}

		//printYell("LookUpInCpLambdaArray() row = %d, col = %d,, b1 = %f, b2 = %f", row, col,b1_curr,b2_curr);

		//return 0.6;
		if (col < 0 || col > 11) {
			//printErr("pitch = %f, col = %d\n",pitch_,col);
			return 0.5;
		}

		//printYell("LookUpInCpLambdaArray() CplArray size = %d", CplArray.Num());

		if (CplArray.Num() < 91) return 0.1;

			Cp1 = CplArray[row][col];
			Cp2 = CplArray[row][col + 1];
			Cpp = Interpolate(pitch_, b1_curr, b2_curr, Cp1, Cp2);



		return Cpp;
	}
	
	double Interpolate(double beta_, double beta1_, double beta2_, double Cp1_, double Cp2_) {
		double ret;
		ret = (Cp2_ - Cp1_) / (beta2_ - beta1_) * (beta_ - beta1_) + Cp1_;
		return ret;
	}

}; // --------------------------------------------------------------------



USTRUCT()
struct FDataRecord_WindT {
	GENERATED_BODY()
		// Actual variables logged depend on specific actor
		double time, theta, omega, windV, tsr, Cp, beta, power;
};

UENUM()
enum WindType {
	steady	    UMETA(DisplayName = "Steady"),
	stepped		UMETA(DisplayName = "Stepped"),
	series	    UMETA(DisplayName = "Time Series"),
};

// -----------------------------------------------------------------------

UCLASS()
class MAS22_SCIENCEPARK_API AMAS22_WindTurbine : public AMAS22_Actor
{
	GENERATED_BODY()


	UPROPERTY(VisibleDefaultsOnly, Category = "Mesh")
		class UCapsuleComponent* CapsuleCollider;
	UPROPERTY(VisibleDefaultsOnly, Category = "Mesh")
		class UStaticMeshComponent* Tower;
	UPROPERTY(VisibleDefaultsOnly, Category = "Mesh")
		class UStaticMeshComponent* Blade1;
	UPROPERTY(VisibleDefaultsOnly, Category = "Mesh")
		class UStaticMeshComponent* Blade2;
	UPROPERTY(VisibleDefaultsOnly, Category = "Mesh")
		class UStaticMeshComponent* Blade3;


public:
	AMAS22_WindTurbine();

private:
	// Parameters used in the cpp ----------------------------------------
	double _R, _rho, _J, _tsrOpt, _cpMax, _region3Kp, _region3Ki, _omegaRated, _region2Beta;
	// Complete above line. Must use underscores

public:
	// Editable variables ------------------------------------------------
	// Complete the following list

	UPROPERTY(EditAnywhere, Config)
		double R = 20;
	UPROPERTY(EditAnywhere, Config)
		double rho = 1.2;
	UPROPERTY(EditAnywhere, Config)
		double J = 613301;
	UPROPERTY(EditAnywhere, Config)
		double omegaRated = 3.861;  // =vRated/tsrOpt/R
	UPROPERTY(EditAnywhere, Config)
		double region3Kp = 2.0;
	UPROPERTY(EditAnywhere, Config)
		double region3Ki = 2.0;
	UPROPERTY(EditAnywhere, Config)
		double region2Beta = 3.7;
	UPROPERTY(EditAnywhere, Config)
		double tsrOpt = 6.6;
	UPROPERTY(EditAnywhere, Config)
		double cpMax = 0.4633;

	UPROPERTY(EditAnywhere, Config, Category = "Wind")
		TEnumAsByte<WindType> windType;
	UPROPERTY(EditAnywhere, Config)
		double vCutIn = 4.0;
	UPROPERTY(EditAnywhere, Config)
		double vRated = 11.7;
	UPROPERTY(EditAnywhere, Config)
		double vCutOut = 25.0;

	UPROPERTY(EditAnywhere, Config, Category = "Wind|Stepped")
		double windStepInterval = 20;
	UPROPERTY(EditAnywhere, Config, Category = "Wind|Stepped")
		double windStepHeight = 1.0;
	UPROPERTY(EditAnywhere, Config, Category = "Wind|TimeSeries")
		double averageWindSpeed = 10.0;
	UPROPERTY(EditAnywhere, Config, Category = "Wind|TimeSeries")
		double changeInterval = 5.0;
	UPROPERTY(EditAnywhere, Config)
		double omegaInit = 2.0;
	UPROPERTY(EditAnywhere, Config)
		double windVInit = 8.0;



	// -------------------------------------------------------------------

protected:

	virtual void Computation();
	virtual void Visualization(float deltaTime);
	virtual void BuildHUDLines();
	virtual void LogData();
	virtual void UserInput(int key);
	virtual void EndPlay(const EEndPlayReason::Type EndPlayReason) override;
	virtual void ApplyConfig() override;

	virtual void Euler(double dT);
	virtual void RKF(double dT);
	virtual void RK(double dT);

	// Overriden functions from MAS22_Actor ---------------------------------

	virtual void ActorInitialize() override;
	virtual void InitializeSolver() override;
	virtual void ComputeDT() override;

	virtual void SetModelParameters() override;
	virtual void UpdateModelParameters() override;
	virtual void GetDefaultParameters() override;

	virtual void SetModelICs() override;
	virtual void UpdateModelICs() override;

	virtual void DumpConsoleAll() override;
	//virtual void InitGlobalLog() override;

	// Console commands specific to this model ------------------------------

	virtual void SetConsoleAddit(FString name, FString strVal) override;

public:

	rhs_WindTurbine* rightHandSide;
	MAS22_Solver_RK* solverRK;
	MAS22_Solver_RKF* solverRKF;

private:
	// Model variables for cpp ---------------------------------------------
	double theta, omega, beta;
	double power;
	double windV;
	double tsr;
	double Cp;
	int region;

	double stepTime;

	// stuff for debugging ------------------
	double currentCp;
	double currentTSR;
	double savedCp;
	double savedTSR;

	//---------------------------------------

	FVector location = { 0,0,0 };  // Will be used in wind farms proc locn

	// Model functions (addit) for cpp -------------------------------------
	double steppedWindSpeed();
	double seriesWindSpeed(double averageSpeed);
	void dumpCplArray(FString message);

	// Adit variables needed by the model ----------------------------------
	//FVector CarStartLocation;


	// TArray for data logging ----------------------------------------------
	TArray<FDataRecord_WindT> dataRecords;

};
